package org.eclipse.albireo.internal;

import java.awt.AWTKeyStroke;
import java.awt.Component;
import java.awt.EventQueue;
import java.awt.Frame;
import java.awt.KeyEventDispatcher;
import java.awt.KeyboardFocusManager;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowFocusListener;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Set;

import javax.swing.text.Caret;
import javax.swing.text.JTextComponent;

import org.eclipse.albireo.core.SwingControl;
import org.eclipse.albireo.core.ThreadingHandler;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;

public class FocusHandler {

    private final Frame frame;
    private final SwingControl swingControl;
    private final Display display;
    private boolean pendingTraverse = false;
    private int extraTabCount = 0;
    private boolean isActiveSwt;
    private boolean isFocusedSwt;
    private static SwingControl activeSwingControl = null;
    
    // Listeners
    private KeyEventDispatcher keyEventDispatcher = new AwtKeyDispatcher();
    private WindowFocusListener awtWindowFocusListener = new AwtWindowFocusListener();
    private FocusListener swtFocusListener = new SwtFocusListener();
    private Listener swtEventFilter = new SwtEventFilter();
    
    // Whether to print debugging information regarding focus events.
    public static final boolean verboseFocusEvents = false;
    public static final boolean verboseKFHEvents = false;
    public static final boolean verboseTraverseOut = false;
    
    static private boolean synthesizeMethodInitialized = false;
    static private Method synthesizeMethod = null;

    public FocusHandler(final SwingControl swingControl, final Frame frame) {
        assert Display.getCurrent() != null;     // On SWT event thread

        if (verboseFocusEvents)
            FocusDebugging.addFocusDebugListeners(swingControl, frame);
        
        this.swingControl = swingControl;
        this.frame = frame;
        display = swingControl.getDisplay();
        
        getSynthesizeMethod(frame.getClass());

        // TODO: optimize by adding just one pair of filters for all SwingControls
        display.addFilter(SWT.Activate, swtEventFilter);
        display.addFilter(SWT.Deactivate, swtEventFilter);
        
        frame.addWindowFocusListener(awtWindowFocusListener);
        KeyboardFocusManager.getCurrentKeyboardFocusManager().addKeyEventDispatcher(keyEventDispatcher);
        
        swingControl.addFocusListener(swtFocusListener);
        
    }
    
    public void dispose() {
        display.removeFilter(SWT.Activate, swtEventFilter);
        display.removeFilter(SWT.Deactivate, swtEventFilter);
        frame.removeWindowFocusListener(awtWindowFocusListener);
        KeyboardFocusManager.getCurrentKeyboardFocusManager().removeKeyEventDispatcher(keyEventDispatcher);
        swingControl.removeFocusListener(swtFocusListener);
    }
    
    // ================ 
    // On a normal change of focus, Swing will turn off any selection
    // in a text field to help indicate focus is lost. This won't happen
    // automatically when transferring to SWT, so turn off the selection
    // manually.
    protected void hideTextSelection() {
        assert EventQueue.isDispatchThread();
        
        Component focusOwner = frame.getMostRecentFocusOwner();
        if (focusOwner instanceof JTextComponent) {
            Caret caret = ((JTextComponent)focusOwner).getCaret();
            if (caret != null) {
                caret.setSelectionVisible(false);
            }
        }
    }
    
    
    // =====
    // Embedded frames in win32 do not support traverse out. For seamless embedding, we
    // check for the need to traverse out here and generate the necessary SWT traversal(s). 
    // TODO: this should be optional
    protected boolean checkForTraverseOut(KeyEvent e) {
        assert EventQueue.isDispatchThread();
        
        // Ignore events outside this frame
        if (frame.isFocused()) {
            Set traverseForwardKeys = frame.getFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS);
            Set traverseBackwardKeys = frame.getFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS);
            AWTKeyStroke key = AWTKeyStroke.getAWTKeyStrokeForEvent(e);
            
            if (!pendingTraverse) {
                // We haven't started to traverse out yet. Check to see if the traversal key has been 
                // hit and that we are at the last/first compoent in the traversal group. 
                
                Component limit = frame.getFocusTraversalPolicy().getLastComponent(frame);
                if (traverseForwardKeys.contains(key) && (limit == e.getComponent() || limit == null)) {
                    // Tabbing forward from last component in frame (or empty frame)
                    pendingTraverse = true;
                    swtTraverse(SWT.TRAVERSE_TAB_NEXT, 1);
                    if (verboseTraverseOut) {
                        trace("AWT: traversing out (forward)");
                    }
                    e.consume();
                    return true;
                }
                
                limit = frame.getFocusTraversalPolicy().getFirstComponent(frame);
                if (traverseBackwardKeys.contains(key) && (limit == e.getComponent() || limit == null)) {
                    // Tabbing backward from first component in frame
                    pendingTraverse = true;
                    swtTraverse(SWT.TRAVERSE_TAB_PREVIOUS, 1);
                    if (verboseTraverseOut) {
                        trace("AWT: traversing out (backward)");
                    }
                    e.consume();
                    return true;
                }
            } else {
                // We received a keystroke while the traverse out is pending. Record any tabs so that the
                // right number of SWT traversals can be done in the SWT thread. This prevents us from losing
                // tabs when typing them very fast. 
                // TODO: this needs to be generalized to all keystrokes, not just traversals
                if (traverseForwardKeys.contains(key)) {
                    if (verboseTraverseOut) {
                        trace("forward traversal typeahead");
                    }
                    extraTabCount++;
                    e.consume();
                    return true;
                } else if (traverseBackwardKeys.contains(key)) {
                    if (verboseTraverseOut) {
                        trace("backward traversal typeahead");
                    }
                    extraTabCount--;
                    e.consume();
                    return true;
                }
            }
        }
        return false;
    }
    
    protected void processTypeAheadKeys() {
        if (pendingTraverse && extraTabCount != 0) {
            if (verboseTraverseOut) {
                trace("Processing typeahead traversals, count=" + extraTabCount);
            }
            int direction = (extraTabCount > 0) ? SWT.TRAVERSE_TAB_NEXT : SWT.TRAVERSE_TAB_PREVIOUS;
            swtTraverse(direction, Math.abs(extraTabCount));
        }
        pendingTraverse = false;
        extraTabCount = 0;
    }

    protected void swtTraverse(final int direction, final int count) {
        assert EventQueue.isDispatchThread();

        ThreadingHandler.getInstance().asyncExec(display, new Runnable() {
            public void run() {
                for(int i = 0; i < count; i++) {
                    doTraverse(direction);
                }
            }

        });
    }

    protected void doTraverse(final int direction) {
        assert Display.getCurrent() != null;

        Control focusControl = display.getFocusControl();
        if (verboseTraverseOut) {
            trace("SWT: traversing, control=" + focusControl);
        }
        if ((focusControl == null) && (activeSwingControl != null)) {
            focusControl = activeSwingControl;
            if (verboseTraverseOut) {
                trace("SWT: current focus control is null; using=" + focusControl);
            }
        }
        if (focusControl != null) {
            boolean traverse = focusControl.traverse(direction);
            if (verboseTraverseOut && !traverse) {
                trace("traverse failed, from=" + focusControl);
            }
        }
    }

    public boolean handleForceFocus(SwingControl swingControl, boolean result) {
        if (Platform.isWin32()) {
            // On Windows, focus queries are unreliable while traversing SwingControls
            // In some cases this causes forceFocus() to incorrectly return false, causing
            // the SwingControl to be skipped while tabbing. Try to fix it here by resetting
            // the forceFocus result back to true when SWT events indicate we really do have 
            // focus. 
            if (!result && isActiveSwt && isFocusedSwt) {
                // Force focus should have returned true
                result = true;
                if (verboseTraverseOut) {
                    trace(" resetting forceFocus return code to true");
                }
            }
        }
        return result;
    }

    // ==============
    // On Windows, when the embedded frame loses focus to an ancestor which then assigns focus to 
    // the embedded composite, the composite receives only a Deactivate event and not an Activate
    // event. This method schedules an async exec, meant to run after the SWT focus change is complete. 
    // It sends the missing activation when necessary, so that the frame can get back focus.
    //
    // This solves the problem where clicking on an RCP view tab causes an embedded Swing control
    // to lose focus. 
    protected void postMissingActivation() {
        assert Platform.isWin32();               // currently only implemented/needed for win32
        
        if (!display.isDisposed()) {
            ThreadingHandler.getInstance().asyncExec(display, new Runnable() {
                public void run() {
                    if (!swingControl.isDisposed() &&
                            
                        // Make sure that this control is in the active shell, so focus is not stolen from other windows    
                        (display.getActiveShell() == swingControl.getShell()) && 
                        
                        // Check that this control still the focus control, despite the recent deactivation.
                        // BUT... Display.getFocusControl is unreliable when another embedded AWT window has 
                        // become active, so to be safe, make sure that no other Swing control has been activated
                        // (otherwise we will steal focus from some other active SwingControl)
                        (swingControl == display.getFocusControl() && (activeSwingControl == null))) {
                        
                        if (verboseFocusEvents) {
                            trace("Manually reactivating " + swingControl);
                        }
                        // Ideally, we would directly activate SwingControl here, but there seems to be 
                        // no way to do that, so activate the underlying AWT embedded frame
                        synthesizeWindowActivation(true);
                    }
                }
            });
        }
    }

    // ==== Fix Eclipse bug 216431 on pre-3.4
    // XEmbeddedFrame does not implement the synthesize method, and the Eclipse bug was fixed only for
    // win32, so we do the same in the methods below
    
    /**
     * This method duplicates the behavior of recent versions of Windows SWT_AWT 
     * when the embedded Composite is activated or deactivated. It is used here to
     * workaround bugs in earlier (pre-3.4) versions of SWT, and to handle cases
     * where the Composite is not properly activated/deactivated, even today. See
     * the callers of this method for more information. 
     * 
     * @param activate <code>true</code> if the embedded frame whould be activated; 
     * <code>false</code> otherwise
     * @return
     */
    protected void synthesizeWindowActivation(final boolean activate) {
        assert Display.getCurrent() != null;     // On SWT event thread
        assert Platform.isWin32();               // Only done on Windows

        EventQueue.invokeLater(new Runnable() {
            public void run() {
                if (synthesizeMethod != null) {
                    // synthesizeWindowActivation() is available. Use it. Normally, this is on
                    // Java 1.5 and higher. 
                    try {
                        if (synthesizeMethod != null) {
                            if (verboseFocusEvents) {
                                trace("Calling synthesizeWindowActivation(" + activate + ")");
                            }
                            synthesizeMethod.invoke(frame,
                                    new Object[] { new Boolean(activate) });
                        }
                    } catch (IllegalAccessException e) {
                        handleSynthesizeException(e);
                    } catch (InvocationTargetException e) {
                        handleSynthesizeException(e);
                    }
                } else {
                    if (activate) {
                        frame.dispatchEvent (new WindowEvent (frame, WindowEvent.WINDOW_ACTIVATED));
                        frame.dispatchEvent (new WindowEvent (frame, WindowEvent.WINDOW_GAINED_FOCUS));
                    } else {
                        frame.dispatchEvent (new WindowEvent (frame, WindowEvent.WINDOW_LOST_FOCUS));
                        frame.dispatchEvent (new WindowEvent (frame, WindowEvent.WINDOW_DEACTIVATED));
                    }
                }
            }
        });
    }

    private void getSynthesizeMethod(Class clazz) {
        if (Platform.isWin32() && !synthesizeMethodInitialized) {
            synthesizeMethodInitialized = true;
            try {
                synthesizeMethod = clazz.getMethod("synthesizeWindowActivation", new Class[]{boolean.class});
			} catch (NoSuchMethodException e) {
                handleSynthesizeException(e);
			}
        }
    }

    private void handleSynthesizeException(Exception e) {
        if (verboseFocusEvents) {
            e.printStackTrace();
        }
    }


    // ========== Listener implementations
    
    protected class SwtEventFilter implements Listener {
        
        public void handleEvent(Event event) {
            if (event.widget != swingControl) {
                return;
            }
            
            switch (event.type) {
            case SWT.Activate:
                isActiveSwt = true;
                activeSwingControl = swingControl;
                if (Platform.isWin32() && (Platform.SWT_VERSION < Platform.SWT_34) && (synthesizeMethod != null)) {
                    synthesizeWindowActivation(true);
                    if (verboseFocusEvents) {
                        trace("Consuming SWT.Activate event: " + event);
                    }
                    event.type = SWT.None;
                }
                
                break;

            case SWT.Deactivate:
                isActiveSwt = false;
                activeSwingControl = null;
                if (Platform.isWin32() && (Platform.SWT_VERSION < Platform.SWT_34) && (synthesizeMethod != null)) {
                    synthesizeWindowActivation(false);
                    if (verboseFocusEvents) {
                        trace("Consuming SWT.Deactivate event: " + event);
                    }
                    event.type = SWT.None;
                }
                if (Platform.isWin32()) {
                    postMissingActivation();
                }
                break;
            }
        }

    }

    protected class SwtFocusListener implements FocusListener {
        public void focusGained(FocusEvent e) {
            isFocusedSwt = true;
        }
        public void focusLost(FocusEvent e) {
            isFocusedSwt = false;
        }
    }

    protected class AwtWindowFocusListener implements WindowFocusListener {
        public void windowGainedFocus(WindowEvent e) {
            assert !pendingTraverse;
            assert extraTabCount == 0;
        }

        public void windowLostFocus(WindowEvent e) {
            if (Platform.isWin32()) {
                hideTextSelection();
                processTypeAheadKeys();
            }
        }
        
    }

    protected class AwtKeyDispatcher implements KeyEventDispatcher {

        public boolean dispatchKeyEvent(KeyEvent e) {
            boolean result = false;
            
            if (Platform.isWin32()) {
                result = checkForTraverseOut(e);
            }
            return result;
        }

    }

    private void trace(String msg) {
        System.err.println(header() + ' ' + msg);
    }
    private String header() {
        return "@" + System.currentTimeMillis() + " " + System.identityHashCode(this);
    }


}
